@@ -6,12 +6,13 @@ import mysql.connector
 from database_infrastructure_local.number_generator import NumberGenerator
 from logger_local.MetaLogger import MetaLogger
 from user_context_remote.user_context import UserContext
+from python_sdk_remote.our_object import OurObject
 
 from .connector import Connector
 from .constants import DEFAULT_SQL_SELECT_LIMIT, LOGGER_CRUD_CODE_OBJECT
 from .cursor import Cursor
 from .table_definition import table_definition
-from .utils import (detect_if_is_test_data, generate_column_name,
+from .utils import (detect_if_is_test_data, generate_id_column_name,
                     generate_table_name, generate_view_name,
                     get_entity_type_by_table_name, get_where_params,
                     process_insert_data_dict, process_update_data_dict, process_upsert_data_dict,
@@ -20,7 +21,10 @@ from .utils import (detect_if_is_test_data, generate_column_name,
                     where_skip_null_values, insert_is_undelete, is_column_in_table,
                     is_end_timestamp_in_table, get_table_columns, group_list_by_columns,
                     process_select_data_dict, generate_where_clause_for_ignore_duplicate,
-                    fix_select_clause_value)
+                    fix_select_clause_value, fix_where_clause_value)
+
+# TODO General question about the overhead of the packages GenericCrud? GenericCrudMl? Mapping? comapring to working directly with the database.
+
 
 # TODO General question about the overhead of the packages GenericCrud? GenericCrudMl? Mapping? comapring to working directly with the database.
 
@@ -40,12 +44,12 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         self.default_schema_name = default_schema_name
         # We do not need a connection per schema, and it makes terrible performance.
         # In special cases, you can always use set_schema.
-        self.connection = Connector.connect(schema_name="INFORMATION_SCHEMA")
+        self.connection = Connector.connect(schema_name=default_schema_name)
         self._cursor = self.connection.cursor()
         self.default_table_name = default_table_name or generate_table_name(default_schema_name)
         self.default_view_table_name = default_view_table_name or generate_view_name(
             self.default_table_name)
-        self.default_column_name = default_column_name or generate_column_name(self.default_table_name)
+        self.default_column_name = default_column_name or generate_id_column_name(self.default_table_name)
         self.default_view_with_deleted_and_test_data = default_view_with_deleted_and_test_data
         self.default_select_clause_value = default_select_clause_value
         self.default_where = default_where
@@ -71,14 +75,11 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             self.logger.warning(f"GenericCRUD.insert({schema_name}.{table_name}) using ignore_duplicate, is it needed?",
                                 object={"data_dict": data_dict})
 
-        # if table_name in table_definition:
-        #     if table_definition[table_name]["is_number_column"]:
-        #         view_name = self._get_view_name(table_name)
-        #         number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
-        #         data_dict["number"] = number
-        # else:
-        #     self.logger.warning(f"database-mysql-local-python generic_crud.py Table {table_name} not found in "
-        #                         f"database-mysql-local.table_definition_table data structure, we might need to run sql2code")
+        # if table_name in table_definition: if table_definition[table_name]["is_number_column"]: view_name =
+        # self._get_view_name(table_name) number = NumberGenerator.get_random_number(schema_name=schema_name,
+        # view_name=view_name) data_dict["number"] = number else: self.logger.warning(f"database-mysql-local-python
+        # generic_crud.py Table {table_name} not found in " f"database-mysql-local.table_definition_table data
+        # structure, we might need to run sql2code")
 
         # TODO: In the future we may want to check this with table_definition
         #   and not with self.is_column_in_table for better performance
@@ -91,6 +92,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                        f"INTO `{schema_name}`.`{table_name}` ({columns}) " \
                        f"VALUES ({values});"
         try:
+            self.logger.info("insert_query=" + insert_query)
             self.cursor.execute(insert_query, params)
             if commit_changes:
                 self.connection.commit()  # TODO: test if it works when schema_name != self.default_schema_name
@@ -192,17 +194,25 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         inserted_rows = self.cursor.get_affected_row_count()
         return inserted_rows
 
+    # TODO Why we change upsert to upsert_with_select_clause? This might break things ...
     def upsert_with_select_clause(self, *, schema_name: str = None, table_name: str = None, view_table_name: str = None,
                                   data_dict: dict = None, where_compare: str = None, params_compare: tuple = None,
                                   data_dict_compare: dict = None, order_by: str = None, compare_with_or: bool = False,
                                   select_clause_value: str = "*") -> dict:
-        """Inserts a new row into the table, or updates an existing row if a row with the
-            same values as data_dict_compare exists,
-            and returns the new row as a dictionary."""
+        """
+        Inserts a new row into the table if a row with the same values does not exist,
+        the logic:
+        1. If data_dict_compare is empty, insert the row and return the id.
+        2. If data_dict_compare is not empty, select the row with the same values as data_dict_compare.
+
+            a. If the row exists, update it with data_dict and return the id.
+
+            b. If the row does not exist, insert it with data_dict and return the id.
+        """
         schema_name = schema_name or self.default_schema_name
         table_name = table_name or self.default_table_name
         view_table_name = view_table_name or self.default_view_table_name
-        column_name = generate_column_name(table_name)
+        column_name = generate_id_column_name(table_name)
         self._validate_args(args=locals())
         result_dict = {}
         if not data_dict:
@@ -213,7 +223,9 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                                              table_name=table_name, data_dict=data_dict)
             data_dict[column_name] = inserted_id
             result_dict = data_dict
-            return result_dict
+            # TODO Changing return inserted_id to result_dict might break things?
+            return result_dict  # upser() was returning inserted_id
+
         where_clause, params = process_upsert_data_dict(data_dict=data_dict_compare, compare_with_or=compare_with_or,
                                                         where_compare=where_compare, params_compare=params_compare)
         # Add table_id if it's not in select_clause_value
@@ -244,6 +256,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                 )
                 raise exception
         result_dict = data_dict
+        result_dict.update({k: v for k, v in row_dict.items() if k not in result_dict}) if row_dict else None
         result_dict[column_name] = table_id
         self.logger.debug(object=locals())
         return result_dict
@@ -253,19 +266,23 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                data_dict: dict = None, where_compare: str = None,
                params_compare: tuple = None, data_dict_compare: dict = None,
                order_by: str = None, compare_with_or: bool = False) -> Optional[int]:
-        """Inserts a new row into the table, or updates an existing row if a row with the
-            same values as data_dict_compare exists,
-            and returns the id of the new row or None if an error occurred."""
+        """
+        Inserts a new row into the table if a row with the same values does not exist,
+        the "same values" are extracted from data_dict_compare.
+        data_dict_compare is a dictionary with the same keys as data_dict, but with the values to compare.
+
+        **you don't need to have the full fields in data_dict_compare, only the fields you want to compare.**
+        """
         table_name = table_name or self.default_table_name
-        column_name = generate_column_name(table_name)
-        row_dict = self.upsert_with_select_clause(
+        column_name = generate_id_column_name(table_name)
+        inserted_id_per_column_dict = self.upsert_with_select_clause(
             schema_name=schema_name, table_name=table_name,
             view_table_name=view_table_name, data_dict=data_dict,
             where_compare=where_compare, params_compare=params_compare,
             data_dict_compare=data_dict_compare, order_by=order_by,
             compare_with_or=compare_with_or, select_clause_value=column_name)
-        table_id = row_dict[column_name]
-        return table_id
+        inserted_id = inserted_id_per_column_dict[column_name]
+        return inserted_id
 
     def _get_existing_duplicate_id(self, schema_name: str, table_name: str, error: Exception,
                                    data_dict: dict) -> int or None:
@@ -316,7 +333,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
     def __get_existing_duplicate_id_with_undelete(
             self, schema_name: str, table_name: str, data_dict: dict) -> int or None:
-        column_name = generate_column_name(table_name)
+        column_name = generate_id_column_name(table_name)
         where, params = self.get_constraint_where_clause(schema_name=schema_name, table_name=table_name,
                                                          data_dict=data_dict)
         row = self.select_one_tuple_by_where(
@@ -338,7 +355,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
     # TODO: test
     def __get_existing_duplicate_id_without_undelete(
             self, *, schema_name: str, table_name: str, data_dict: dict) -> int or None:
-        column_name = generate_column_name(table_name)
+        column_name = generate_id_column_name(table_name)
         where, params = self.get_constraint_where_clause(schema_name=schema_name, table_name=table_name,
                                                          data_dict=data_dict)
         if is_end_timestamp_in_table(table_name=table_name):
@@ -458,11 +475,18 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         if not where:
             raise Exception("update_by_where requires a 'where'")
 
-        update_query = f"UPDATE `{schema_name}`.`{table_name}` " \
-                       f"SET {set_values} updated_timestamp=CURRENT_TIMESTAMP() " \
-                       f"WHERE {where} " + \
-                       (f"ORDER BY {order_by} " if order_by else "") + \
-                       f"LIMIT {limit};"
+        if 'mapping' in self.default_table_name:
+            update_query = f"UPDATE `{schema_name}`.`{table_name}` " \
+                           f"SET {set_values} " \
+                           f"WHERE {where} " + \
+                           (f"ORDER BY {order_by} " if order_by else "") + \
+                           f"LIMIT {limit};"
+        else:
+            update_query = f"UPDATE `{schema_name}`.`{table_name}` " \
+                           f"SET {set_values} updated_timestamp=CURRENT_TIMESTAMP() " \
+                           f"WHERE {where} " + \
+                           (f"ORDER BY {order_by} " if order_by else "") + \
+                           f"LIMIT {limit};"
         where_params = params or tuple()
 
         self.cursor.execute(update_query, data_dict_params + where_params)
@@ -520,6 +544,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         select_clause_value = select_clause_value or self.default_select_clause_value
         select_clause_value = fix_select_clause_value(select_clause_value=select_clause_value)
         where = where or self.default_where
+        where = fix_where_clause_value(where_clause_value=where)
         where = self.__where_security(where=where, view_name=view_table_name)
         self._validate_args(args=locals())
 
@@ -531,7 +556,14 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             # once done, we prefer not to send it as parameter, to allow changing the name later without changing the code.
             # In the future we will change to default_view_with_test_data that does not show deleted data, so that
             # we can select-delete-select in a similar way and verify the deletion.
-            # Will we ever have a case where there will be a view with no assosiated table that we will need to see deleted data in there? It appears no.
+            # Will we ever have a case where there will be a view with no assosiated table that we will need to see deleted data in there? It appears no.  # noqa E501
+
+            # TODO: auto detect with entity table (save in memory first) once done, we prefer not to send it as
+            #  parameter, to allow changing the name later without changing the code. In the future we will change to
+            #  default_view_with_test_data that does not show deleted data, so that we can select-delete-select in a
+            #  similar way and verify the deletion. Will we ever have a case where there will be a view with no
+            #  assosiated table that we will need to see deleted data in there? It appears no.
+
             if self.default_view_with_deleted_and_test_data:
                 view_table_name = self.default_view_with_deleted_and_test_data
             else:
@@ -606,8 +638,10 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         result = self.convert_to_dict(row=result, select_clause_value=select_clause_value)
         return result
 
+    # TODO Why we need to send "select_clause_value" if we provided "default_select_clause_value" in the constructor. Added "= None" to try and resolve it.
+    # TODO Add ASC/DESC parameter to all methods which have order_by parameter
     def select_one_value_by_column_and_value(
-            self, *, select_clause_value: str, schema_name: str = None, view_table_name: str = None,
+            self, *, select_clause_value: str = None, schema_name: str = None, view_table_name: str = None,
             column_name: str = None, column_value: Any,
             distinct: bool = False, order_by: str = None, skip_null_values: bool = True) -> Any:
         """Selects one value from the table by ID and returns it."""
@@ -743,7 +777,8 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             self, *, select_clause_value: str, schema_name: str = None, view_table_name: str = None,
             data_dict: dict, distinct: bool = False, limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None,
             skip_null_values: bool = True) -> tuple:
-        """Selects multiple values from the table based on a WHERE clause from the data_dict and returns them as a list."""
+        """Selects multiple values from the table based on a WHERE clause from the data_dict and returns them as a
+        list."""
         select_clause_value = select_clause_value or self.default_select_clause_value
         validate_single_clause_value(select_clause_value)
         where, params = process_select_data_dict(data_dict)
@@ -1022,7 +1057,9 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         select_clause_value = select_clause_value or entity_name + "_id"
         test_entity_id = self.select_one_value_by_column_and_value(
             schema_name=schema_name or self.default_schema_name, view_table_name=view_name,
-            column_name='is_test_data', column_value='1', select_clause_value=select_clause_value)
+            column_name='is_test_data', column_value='1', select_clause_value=select_clause_value
+            # TODO Make is implicit ASC 
+            , order_by=select_clause_value ) # To bring alway the 1st is_test_data so the test can relay on the value.
         if not test_entity_id:  # TODO: test
             insert_kwargs = insert_kwargs or {}
             create_kwargs = create_kwargs or {}
@@ -1034,10 +1071,42 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                 test_entity_id = insert_function(**insert_kwargs)
         self.logger.debug(object=locals())
         return test_entity_id
+    
+    def create_view(self, schema_name=None, table_name=None, view_name=None):
+        if table_name is not None and '_view' in table_name:
+            return
+        if schema_name is None:
+            schema_name = self.default_schema_name
+
+        if table_name is None:
+            table_name = self.default_table_name
+
+        if view_name is None:
+            view_name = self.default_view_table_name
+
+        print(f"Creating view {view_name} from {table_name}")
+        create_view_query = f"""
+        CREATE OR REPLACE VIEW `{schema_name}`.`{view_name}` AS
+        SELECT * FROM `{schema_name}`.`{table_name}`;
+        """
+
+        self.cursor.execute(create_view_query)
+
+    def create_column(self, schema_name: str, table_name: str, column_name:str, data_type: str, after_column:str = None, default_value: str = None):
+        query = 'ALTER TABLE ' + schema_name + "." + table_name + " ADD COLUMN " + column_name + " " + data_type + " NULL"
+        if default_value is not None:
+            query += " DEFAULT " + str(default_value)
+        if after_column:
+            query += " AFTER " + after_column
+        self.cursor.execute(query)
 
     def merge_entities(self, entity_id1: int, entity_id2: int):
         table_name = self.default_table_name
-        table_prefix = table_name[:-6]
+        if 'ml' in table_name:
+            table_prefix = table_name.replace("_ml_table", "")
+        else:
+            table_prefix = table_name.replace("_table", "")
+            table_prefix = table_name[:-6]
 
         # establish which id is being merged/ended and which one it is being merged into
         end_id = entity_id1
@@ -1048,9 +1117,45 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         new_id = main_id
 
         self.update_by_column_and_value(
-            schema_name= self.default_schema_name,
-            table_name= self.default_table_name,
-            column_name= self.default_column_name,
+            schema_name=self.default_schema_name,
+            table_name=self.default_table_name,
+            column_name=self.default_column_name,
             column_value=old_id,
-            data_dict={f'new_{table_prefix}_id' : new_id},
+            data_dict={f'new_{table_prefix}_id': new_id},
         )
+
+    def foreach(self, where, limit, function, id_column_name, select_function: callable = None, **kwargs):
+        # TODO: method select_multi_value_by_where() is not working as expected
+
+        # result = self.select_multi_dict_by_where(where=where, limit=limit, select_clause_value="*")  # select_clause_value="test_mysql_id, number, name") # noqa
+        select_function = select_function or self.select_multi_dict_by_where
+        result = select_function(where=where, limit=limit, select_clause_value="*")  # select_clause_value="test_mysql_id, number, name") # noqa
+
+        # self, *, schema_name: str = None, view_table_name: str = None, select_clause_value: str = None,
+        # where: str = None, params: tuple = None, distinct: bool = False, limit: int = DEFAULT_SQL_SELECT_LIMIT,
+        # order_by: str = None, skip_null_values: bool = True) -> tuple:
+        responses_dict = {}
+        for r in result:
+            # print(f"normal r:{r}")
+            r_our = OurObject(id_column_name=id_column_name, **r)
+            kwargs["object"] = r_our
+
+            # print(f"our r:{r_our}")
+            # response = function(r_our)
+            response = function(**kwargs)
+            responses_dict[id_column_name] = response
+        return responses_dict
+
+    def get_view_name(self, schema_name: str) -> str:
+
+        if self.default_view_with_deleted_and_test_data:
+            view_table_name = self.default_view_with_deleted_and_test_data
+        else:
+            if schema_name == "test":
+                view_table_name = f"{schema_name}_mysql_general_view"
+            else:
+                view_table_name = f"{schema_name}_view"
+
+            view_table_name = replace_view_with_table(view_table_name=view_table_name)  # select_clause_value=select_clause_value)
+
+        return view_table_name
