@@ -2,15 +2,16 @@ import datetime
 import re
 from functools import lru_cache
 from typing import Optional
-
+from typing import Tuple
 from language_remote.lang_code import LangCode
 from logger_local.MetaLogger import MetaLogger
 from user_context_remote.user_context import UserContext
 
+# PyCham on MacOS need . before
 from .constants import DEFAULT_SQL_SELECT_LIMIT, LOGGER_CRUD_ML_CODE_OBJECT
 from .generic_crud import GenericCRUD
 from .table_definition import table_definition
-from .utils import generate_column_name, generate_view_name, is_column_in_table
+from .utils import generate_id_column_name, generate_view_name, is_column_in_table
 
 IS_MAIN_COLUMN_NAME = "is_main"
 
@@ -28,7 +29,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
                  default_table_name: str = None, default_view_table_name: str = None,
                  default_ml_table_name: str = None, default_ml_view_table_name: str = None,
                  default_view_with_deleted_and_test_data: str = None,
-                 default_column_name: str = None,
+                 default_column_name: str = None,  # default_select_clause_value: str = "*"
                  is_main_column_name: str = IS_MAIN_COLUMN_NAME, is_test_data: bool = False) -> None:
         """Initializes the GenericCRUDML class. If connection is not provided,
         a new connection will be created."""
@@ -36,6 +37,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         self.default_table_name = default_table_name or (default_ml_table_name or "").replace("_ml", "")
         self.default_ml_table_name = default_ml_table_name or self.generate_ml_table_name(default_table_name)
         self.default_ml_view_table_name = default_ml_view_table_name or generate_view_name(self.default_ml_table_name)
+        self.default_view_with_deleted_and_test_data = default_view_with_deleted_and_test_data
         self.logger.info("GenericCRUDML.__init__", object={
             "default_ml_table_name": self.default_ml_table_name,
             "default_ml_view_table_name": self.default_ml_view_table_name})
@@ -70,14 +72,14 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
     def add_value(self, *, schema_name: str = None, data_dict: dict = None, data_ml_dict: dict,
                   column_name: str = None,
                   table_id: int = None, lang_code: LangCode = None,
-                  is_main: bool or None = None, table_name: str = None,
+                  is_main: bool | None = None, table_name: str = None,
                   ml_table_name: str = None) -> Optional[tuple]:
         schema_name = schema_name or self.default_schema_name
 
         lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_dict=data_ml_dict)
         table_name = table_name or self.default_table_name
         # TODO Shall we also add "or self.default_column_name"?
-        column_name = column_name or generate_column_name(table_name)
+        column_name = column_name or generate_id_column_name(table_name)
         ml_view_name = generate_view_name(ml_table_name) if ml_table_name else self.default_ml_view_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
 
@@ -133,28 +135,32 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
             self.cursor = old_cursor
         return table_id, ml_table_id
 
-    # TODO add parameter if_same_even_deleted_return_existing_id=true - When inserting a deleted entity (all entities except person and tables of codes i.e. country, state, county, city, neighbourhood, street ...) and table_definition.insert_is_undelete=false, we should null all the unique fields of the deleted entity
+    # TODO add parameter if_same_even_deleted_return_existing_id=true - When inserting a deleted entity (all entities except person and tables of codes i.e. country, state, county, city, neighbourhood, street ...) and table_definition.insert_is_undelete=false, we should null all the unique fields of the deleted entity  # noqa
     def add_value_if_not_exist(
             self, *, data_dict: dict = None, data_ml_dict: dict,
             table_id: int = None, lang_code: LangCode = None,
             is_main: bool = None, table_name: str = None, ml_table_name: str = None, schema_name: str = None,
-            order_by: str = None) -> (int, int):
+            order_by: str = None) -> Tuple[int, int]:
 
         schema_name = schema_name or self.default_schema_name
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
-        column_name = generate_column_name(table_name)
-        ml_column_name = generate_column_name(ml_table_name)
+        column_name = generate_id_column_name(table_name)
+        ml_column_name = generate_id_column_name(ml_table_name)
 
         # check if the row exists in the ml_table
-        name = data_ml_dict.get("name") or data_ml_dict.get("title")
-        title = data_ml_dict.get("title") or data_ml_dict.get("name")
+        if data_dict:
+            name = data_dict.get("name") or data_dict.get("title")
+        else:
+            name = data_ml_dict.get("name") or data_ml_dict.get("title")
+        # name = data_dict.get("name") or data_ml_dict.get("title")
+        title = data_ml_dict.get("title") or data_dict.get("name")
         if not name:
             raise ValueError("name or title is required for add_value_if_not_exist")
         table_id = table_id or self.get_id_by_name(
             schema_name=schema_name, table_name=table_name, ml_table_name=ml_table_name, name=name,
             lang_code=lang_code, column_name=column_name, order_by=order_by)
-        ml_table_id = self.get_ml_id_by_name(
+        ml_table_id = self.get_ml_id_by_title(
             schema_name=schema_name, ml_table_name=ml_table_name, name=title, lang_code=lang_code,
             column_name=ml_column_name, order_by=order_by)
         if not table_id:
@@ -168,13 +174,13 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
             self, *, data_dict: dict = None, data_ml_dict: dict,
             ml_table_id: int, table_id: int, lang_code: LangCode = None,
             is_main: bool = None, table_name: str = None, ml_table_name: str = None, schema_name: str = None,
-            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> (int, int):
+            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> Tuple[int, int]:
 
         schema_name = schema_name or self.default_schema_name
         lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_dict=data_ml_dict)
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
-        column_name = generate_column_name(table_name)
+        column_name = generate_id_column_name(table_name)
 
         if table_id is None:
             raise ValueError("table_id is required for update_value")
@@ -197,7 +203,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
 
             data_ml_dict[column_name] = table_id
             data_ml_dict["lang_code"] = lang_code_str
-            column_name = generate_column_name(ml_table_name)
+            column_name = generate_id_column_name(ml_table_name)
 
             super().update_by_column_and_value(schema_name=schema_name, table_name=ml_table_name,
                                                data_dict=data_ml_dict,
@@ -218,18 +224,18 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
             self, *, data_dict: dict = None, data_ml_dict: dict, data_dict_compare: dict = None,
             table_id: int = None, lang_code: LangCode = None, is_main: bool = None, table_name: str = None,
             ml_table_name: str = None, schema_name: str = None, compare_view_name: str = None,
-            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> (int, int):
+            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> Tuple[int, int]:
         compare_view_name = compare_view_name or self.default_ml_view_table_name
 
         schema_name = schema_name or self.default_schema_name
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
-        column_name = generate_column_name(table_name)
-        ml_column_name = generate_column_name(ml_table_name)
+        column_name = generate_id_column_name(table_name)
+        ml_column_name = generate_id_column_name(ml_table_name)
 
         # check if the row exists in the ml_table
-        name = data_ml_dict.get("name") or data_ml_dict.get("title")
-        title = data_ml_dict.get("title") or data_ml_dict.get("name")
+        name = data_dict.get("name") or data_ml_dict.get("title")
+        title = data_ml_dict.get("title") or data_dict.get("name")
         name_compare = name
         title_to_compare = title
         if not name:
@@ -242,7 +248,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
             schema_name=schema_name, table_name=table_name, ml_table_name=ml_table_name,
             name=name_compare, compare_view_name=compare_view_name,
             column_name=column_name, order_by=order_by)
-        ml_table_id = self.get_ml_id_by_name(
+        ml_table_id = self.get_ml_id_by_title(
             schema_name=schema_name, ml_table_name=ml_table_name, name=title_to_compare, lang_code=lang_code,
             compare_view_name=compare_view_name, column_name=ml_column_name, order_by=order_by)
         if not table_id or not ml_table_id:
@@ -262,16 +268,16 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
 
             data_dict_compare: dict = None, table_id: int = None, order_by: str = None,
             table_name: str = None, ml_table_name: str = None, schema_name: str = None, lang_code: LangCode = None,
-            compare_view_name: str = None, limit: int = DEFAULT_SQL_SELECT_LIMIT) -> (int, list[int]):
+            compare_view_name: str = None, limit: int = DEFAULT_SQL_SELECT_LIMIT) -> Tuple[int, list[int]]:
         compare_view_name = compare_view_name or self.default_ml_view_table_name
         schema_name = schema_name or self.default_schema_name
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
-        column_name = generate_column_name(table_name)
-        ml_column_name = generate_column_name(ml_table_name)
+        column_name = generate_id_column_name(table_name)
+        ml_column_name = generate_id_column_name(ml_table_name)
 
         # check if the row exists in the ml_table
-        name_and_abbreviations = data_ml_dict.get("title") or data_ml_dict.get("name")
+        name_and_abbreviations = data_ml_dict.get("title") or data_dict.get("name")
         if not name_and_abbreviations:
             raise ValueError("name or title is required for upsert_value")
         if data_dict_compare:
@@ -295,7 +301,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
             schema_name=schema_name, table_name=table_name, ml_table_name=ml_table_name,
             name=name_compare, lang_code=lang_code, compare_view_name=compare_view_name,
             column_name=column_name, order_by=order_by)
-        ml_table_id = self.get_ml_id_by_name(
+        ml_table_id = self.get_ml_id_by_title(
             schema_name=schema_name, ml_table_name=ml_table_name, name=name_compare, lang_code=lang_code,
             compare_view_name=compare_view_name, column_name=ml_column_name, order_by=order_by)
         ml_ids_list = []
@@ -339,7 +345,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
 
     def get_values_tuple(self, *, table_id: int, schema_name: str = None, ml_view_table_name: str = None,
                          lang_code: LangCode = None, column_name: str = None,
-                         select_clause_value: str = "*", order_by: str = None) -> (int, int):
+                         select_clause_value: str = "*", order_by: str = None) -> Tuple[int, int]:
         lang_code_str = self._get_lang_code_str(lang_code=lang_code)
         ml_view_table_name = ml_view_table_name or self.default_ml_view_table_name
         column_name = column_name or self.default_column_name
@@ -355,7 +361,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
                         column_name: str = None, order_by: str = None) -> dict:
 
         lang_code_str = self._get_lang_code_str(lang_code=lang_code)
-        column_name = column_name or generate_column_name(self.default_table_name)
+        column_name = column_name or generate_id_column_name(self.default_table_name)
         result = super().select_one_dict_by_where(
             view_table_name=self.default_ml_view_table_name, order_by=order_by,
             where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str))
@@ -363,7 +369,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         return result
 
     def get_values_dict_list(self, *, table_id: int, lang_code: LangCode = None,
-                             column_name: str = None, order_by: str = None, schema_name: str = None) -> list:
+                             column_name: str = None, order_by: str = None, schema_name: str = None, limit: int = DEFAULT_SQL_SELECT_LIMIT) -> list:
         schema_name = schema_name or self.default_schema_name
 
         column_name = column_name or self.default_column_name
@@ -372,15 +378,30 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         result = self.select_multi_dict_by_where(
             schema_name=schema_name, view_table_name=self.default_ml_view_table_name,
             where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str),
-            order_by=order_by)
+            order_by=order_by, limit=limit)
+        return result
+
+    def get_values_dict_list_by_where(self, *, lang_code: LangCode = None,  # table_id: int,
+                                      column_name: str = None, order_by: str = None, schema_name: str = None,
+                                      limit: int = DEFAULT_SQL_SELECT_LIMIT, where: str, select_clause_value: str) -> list:
+        schema_name = schema_name or self.default_schema_name
+
+        view_table_name = self.get_view_name(schema_name=schema_name)
+
+        column_name = column_name or self.default_column_name
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
+
+        result = self.select_multi_dict_by_where(
+            schema_name=schema_name, view_table_name=view_table_name, select_clause_value=select_clause_value,
+            where=f"{where} AND lang_code=%s", params=(lang_code_str,),
+            order_by=order_by, limit=limit)
         return result
 
     def get_main_values_tuple(self, *, table_id: int, column_name: str = None,
-                              select_clause_value: str = "*", order_by: str = None, schema_name: str = None) -> (
-            int, int):
+                              select_clause_value: str = "*", order_by: str = None, schema_name: str = None) -> Tuple[int, int]:
         schema_name = schema_name or self.default_schema_name
 
-        column_name = column_name or generate_column_name(self.default_table_name)
+        column_name = column_name or generate_id_column_name(self.default_table_name)
         result = self.select_one_tuple_by_where(
             schema_name=schema_name, view_table_name=self.default_ml_view_table_name,
             where=f"{column_name}=%s AND is_main=True", params=(table_id,),
@@ -410,7 +431,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
         compare_view_name = compare_view_name or generate_view_name(ml_table_name)
-        column_name = column_name or generate_column_name(table_name)
+        column_name = column_name or generate_id_column_name(table_name)
         if is_column_in_table(table_name=ml_table_name, column_name="title"):
             where = "`title`=%s"
             params = (name,)
@@ -429,16 +450,16 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
             select_clause_value=column_name, order_by=order_by)
         return result
 
-    def get_ml_id_by_name(self, *, name: str, schema_name: str = None, ml_table_name: str = None,
-                          lang_code: LangCode = None,
-                          column_name: str = None, order_by: str = None,
-                          compare_view_name: str = None) -> Optional[int]:
+    def get_ml_id_by_title(self, *, name: str, schema_name: str = None, ml_table_name: str = None,
+                           lang_code: LangCode = None,
+                           column_name: str = None, order_by: str = None,
+                           compare_view_name: str = None) -> Optional[int]:
 
         lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
         ml_table_name = ml_table_name or self.default_ml_table_name
         schema_name = schema_name or self.default_schema_name
         compare_view_name = compare_view_name or generate_view_name(ml_table_name)
-        ml_column_name = column_name or generate_column_name(ml_table_name)
+        ml_column_name = column_name or generate_id_column_name(ml_table_name)
         if is_column_in_table(table_name=ml_table_name, column_name="title"):
             where = "title=%s"
             params = (name,)
@@ -471,15 +492,15 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         return result
 
     def get_all_ml_ids_by_id(self, table_id: int, ml_table_name: str = None,
-                         column_name: str = None, ml_column_name: str = None,
-                         order_by: str = None, compare_view_name: str = None, schema_name: str = None) -> tuple:
+                             column_name: str = None, ml_column_name: str = None,
+                             order_by: str = None, compare_view_name: str = None, schema_name: str = None) -> tuple:
         schema_name = schema_name or self.default_schema_name
 
         compare_view_name = compare_view_name or generate_view_name(ml_table_name)
         result = super().select_multi_value_by_where(
             schema_name=schema_name, view_table_name=compare_view_name,
             select_clause_value=ml_column_name, order_by=order_by,
-            where=f"{column_name}=%s" , params=(table_id))
+            where=f"{column_name}=%s", params=(table_id))
         return result
 
     def delete_by_name(self, *, name: str, lang_code: LangCode = None) -> int:
@@ -502,7 +523,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
     def _update_old_main_value_to_false(self, *, schema_name: str, table_id: int, table_name: str,
                                         ml_table_name: str) -> int:
         data_ml_dict = {self.is_main_column_name: False}
-        column_name = generate_column_name(table_name)
+        column_name = generate_id_column_name(table_name)
         where = f"{column_name}=%s AND {self.is_main_column_name} = True"
         updated_id = super().update_by_where(schema_name=schema_name, table_name=ml_table_name,
                                              data_dict=data_ml_dict, where=where, params=(table_id,))
@@ -526,11 +547,16 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         DEFAULT_ORDER_BY = "updated_timestamp DESC"
         schema_name = self.default_schema_name
         table_name = self.default_table_name
-        table_prefix = table_name[:-6]
+        table_prefix = table_name.replace("_ml_table", "")
         generated_ml_table_name = table_prefix + '_ml_table'
         generated_ml_view_name = table_prefix + '_ml_view'
         generated_table_name = table_prefix + '_table'
-        gcrml = GenericCRUDML(default_schema_name=schema_name, default_ml_table_name=generated_ml_table_name,is_test_data=True, default_table_name=generated_table_name, default_ml_view_table_name= generated_ml_view_name)
+        gcrml = GenericCRUDML(default_schema_name=schema_name,
+                              default_ml_table_name=generated_ml_table_name,
+                              is_test_data=True,
+                              default_table_name=generated_table_name,
+                              default_ml_view_table_name=generated_ml_view_name,
+                              is_main_column_name='is_main_title')
         # establish which id is being merged/ended and which one it is being merged into
         end_id = entity_id1
         main_id = entity_id2
@@ -538,37 +564,63 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         # Data to update
         old_id = end_id
         new_id = main_id
-        #get the ml ids of the city id which is being merged
+        # get the ml ids of the city id which is being merged
 
         generated_ml_column_name = table_prefix + '_ml_id'
         generated_column_name = table_prefix + '_id'
-        ml_ids = gcrml.get_all_ml_ids_by_id(table_id=old_id,ml_column_name=generated_ml_column_name,order_by=DEFAULT_ORDER_BY, ml_table_name=generated_ml_table_name, column_name=generated_column_name, compare_view_name=generated_ml_view_name)
+        ml_ids = gcrml.get_all_ml_ids_by_id(table_id=old_id, ml_column_name=generated_ml_column_name, order_by=DEFAULT_ORDER_BY,
+                                            ml_table_name=generated_ml_table_name,
+                                            column_name=generated_column_name, compare_view_name=generated_ml_view_name)
 
+        # get a list of existing langcodes for the id which is being merged into
+        langcode_list = self.get_lang_codes_by_table_id(table_id=new_id, ml_view_table_name=generated_ml_view_name,
+                                                        schema_name=self.default_schema_name)
+
+        for id in ml_ids:
+            # get the langcode of the ml id being merged. if it already exists in the id being merged into, then set this one to false.
+            langcode = gcrml.select_one_value_by_column_and_value(select_clause_value='lang_code', column_name=generated_ml_column_name,
+                                                                  column_value=id, view_table_name=generated_ml_view_name)
+            if langcode in langcode_list:
+                gcrml.update_by_column_and_value(
+                    table_name=generated_ml_table_name,
+                    column_name=generated_ml_column_name,
+                    column_value=id,
+                    data_dict={'is_main_title': None}, )
+            # update the id for the ml_id to the new id
+            gcrml.update_by_column_and_value(
+                table_name=generated_ml_table_name,
+                column_name=generated_ml_column_name,
+                column_value=id,
+                data_dict={generated_column_name: new_id, }
+            )
+
+        gcrml.update_by_column_and_value(
+            table_name=generated_table_name,
+            column_name=generated_column_name,
+            column_value=old_id,
+            data_dict={"end_timestamp": datetime.datetime.now()},
+        )
         # Set the main id
         if main_entity_ml_id is not None:
             gcrml.update_by_column_and_value(
                 table_name=generated_ml_table_name,
                 column_name=generated_ml_column_name,
                 column_value=main_entity_ml_id,
-                data_dict={'is_main': True },)
-        for id in ml_ids:
-            gcrml.select_one_value_by_column_and_value(
-                select_clause_value = 'title',
-                schema_name=schema_name,
-                view_table_name=generated_ml_view_name,
-                column_name=generated_ml_column_name,
-                column_value=id,
-            )
-            gcrml.update_by_column_and_value(
-                table_name=generated_ml_table_name,
-                column_name=generated_ml_column_name,
-                column_value=id,
-                data_dict={generated_column_name: new_id,}
-            )
+                data_dict={'is_main_title': True},)
 
-        gcrml.update_by_column_and_value(
-        table_name=generated_table_name,
-        column_name=generated_column_name,
-        column_value=old_id,
-        data_dict={"end_timestamp": datetime.datetime.now()},
-    )
+    def get_lang_codes_by_table_id(self, table_id: int, schema_name: str = None, ml_view_table_name: str = None) -> list:
+        schema_name = schema_name or self.default_schema_name
+        ml_view_table_name = ml_view_table_name or self.default_ml_view_table_name
+        table_name = self.default_table_name
+        table_prefix = table_name.replace("_ml_table", "")
+        generated_column_name = table_prefix + '_id'
+
+        result = self.select_multi_dict_by_where(
+            schema_name=schema_name,
+            view_table_name=ml_view_table_name,
+            where=f"{generated_column_name}=%s",
+            params=(table_id,),
+            select_clause_value="lang_code"
+        )
+
+        return [entry['lang_code'] for entry in result if 'lang_code' in entry]
