@@ -1,6 +1,9 @@
 from .to_sql_interface import ToSQLInterface
 
 
+# TODO: rename to OurPoint
+# TODO We should consier to change is to OurPoint everywere
+# Classes which uses Point Class: Location, Organization, ContactGroup
 class Point(ToSQLInterface):
     def __init__(self, longitude: float, latitude: float) -> None:
         self.longitude = longitude
@@ -20,4 +23,7 @@ class Point(ToSQLInterface):
                 self.latitude == other_point.latitude)
 
     def __repr__(self) -> str:
-        return f"Point(longitude={self.longitude}, latitude={self.latitude})"
+        return f"OurPoint(longitude={self.longitude}, latitude={self.latitude})"
+
+
+ToSQLInterface.register(Point)
