@@ -2,7 +2,7 @@ import copy
 import inspect
 import os
 from functools import lru_cache
-from typing import Any, Optional
+from typing import Any, Optional, Tuple
 
 from logger_local.LoggerLocal import Logger
 from logger_local.MetaLogger import module_wrapper
@@ -73,7 +73,8 @@ def validate_none_select_table_name(table_name: str) -> None:
             f"Table name must end with '_table' in this environment  (got {table_name})")
 
 
-def process_insert_data_dict(data_dict: dict or None) -> (str, str, tuple):
+# TODO Shall we change it from a function to a method in GenericCrud
+def process_insert_data_dict(data_dict: dict | None) -> Tuple[str, str, tuple]:
     """Example:
     Input: {"name": "John", "coordinate": Point(1, 2)}
     Output: ("`name`, `coordinate`",
@@ -81,7 +82,7 @@ def process_insert_data_dict(data_dict: dict or None) -> (str, str, tuple):
             ("John", ))
     """
     if not data_dict:
-        columns_str, values_str, params = '', '', {}
+        columns_str, values_str, params = '', '', ()
         return columns_str, values_str, params
 
     columns = []
@@ -89,8 +90,17 @@ def process_insert_data_dict(data_dict: dict or None) -> (str, str, tuple):
     params = tuple()
 
     for key, value in data_dict.items():
-        columns.append(f"`{key}`")
-        if isinstance(value, ToSQLInterface):
+        # new_key = f'`{key}`' if '.' in key else key
+        if key.startswith("`") and key.endswith("`"):
+            new_key = key
+        else:
+            new_key = f"`{key}`"
+
+        columns.append(new_key)
+
+        is_instance = isinstance(value, ToSQLInterface)
+
+        if is_instance:
             values.append(value.to_sql())
         else:
             values.append('%s')
@@ -102,9 +112,10 @@ def process_insert_data_dict(data_dict: dict or None) -> (str, str, tuple):
     return columns_str, values_str, params
 
 
-def process_upsert_data_dict(data_dict: dict or None, compare_with_or: bool,
-                             where_compare: str or None = None,
-                             params_compare: tuple or None = None) -> (str, tuple):
+def process_upsert_data_dict(data_dict: dict | None, compare_with_or: bool,
+                             where_compare: str | None = None,
+                             params_compare: tuple | None = None) -> Tuple[str, tuple]:
+
     where_clauses = []
     params = []
     for column, value in data_dict.items():
@@ -130,7 +141,7 @@ def process_upsert_data_dict(data_dict: dict or None, compare_with_or: bool,
     return where_clause, params
 
 
-def process_update_data_dict(data_dict: dict or None) -> (str, tuple):
+def process_update_data_dict(data_dict: dict | None) -> Tuple[str, tuple]:
     """Example:
     Input: {"name": "John", "coordinate": Point(1, 2)}
     Output: ("name=%s, coordinate=Point(1, 2)", ("John", ))
@@ -153,7 +164,7 @@ def process_update_data_dict(data_dict: dict or None) -> (str, tuple):
     return set_values_str, params
 
 
-def process_select_data_dict(data_dict: dict or None, select_with_or: bool = False) -> (str, tuple):
+def process_select_data_dict(data_dict: dict | None, select_with_or: bool = False) -> Tuple[str, tuple]:
     """Example:
     Input: {"name": "John", "coordinate": Point(1, 2)}
     Output: ("name=%s AND coordinate=Point(1, 2)", ("John", ))
@@ -190,7 +201,7 @@ def detect_if_is_test_data() -> bool:
     return is_test_data
 
 
-def get_entity_type_by_table_name(table_name: str) -> int or None:
+def get_entity_type_by_table_name(table_name: str) -> int | None:
     """Returns the entity_type_id of the table."""
     entity_type_id = table_definition.get(table_name, {}).get("entity_type_id1")
     return entity_type_id
@@ -210,7 +221,7 @@ def generate_view_name(table_name: Optional[str]) -> Optional[str]:
     return view_name
 
 
-def generate_column_name(table_name: Optional[str]) -> Optional[str]:
+def generate_id_column_name(table_name: Optional[str]) -> Optional[str]:
     if table_name:
         column_name = table_name.replace("_table", "_id")
         return column_name
@@ -241,7 +252,7 @@ def get_where_params(column_name: str, column_value: Any) -> tuple:
     return where, params
 
 
-def where_skip_null_values(where: str or None, select_clause_value: str,
+def where_skip_null_values(where: str | None, select_clause_value: str,
                            skip_null_values: bool = True) -> str:
     if skip_null_values:
         validate_single_clause_value(select_clause_value)
@@ -308,7 +319,7 @@ def get_table_columns(table_name: str = None) -> tuple:
     return table_columns_tuple
 
 
-def group_list_by_columns(list_of_dicts: list, group_by: str, condense: bool = True) -> dict[tuple or str, list]:
+def group_list_by_columns(list_of_dicts: list, group_by: str, condense: bool = True) -> dict[tuple | str, list]:
     """if condense is True and there is only one column left in the dict, return the value instead of the dict.
     Examples:
     Input: [{"name": "John", "age": 25}, {"name": "John", "age": 26}], group_by="name"
@@ -328,7 +339,7 @@ def group_list_by_columns(list_of_dicts: list, group_by: str, condense: bool = T
     if any(col not in list_of_dicts[0] for col in group_by):
         raise ValueError(f"{group_by} should be a subset of {tuple(list_of_dicts[0].keys())}")
     if len(group_by) == len(list_of_dicts[0]):
-        raise ValueError(f"Column names in group_by must be less than the number of columns in the list of dicts")
+        raise ValueError("Column names in group_by must be less than the number of columns in the list of dicts")
 
     for dict_row in list_of_dicts:
         key = tuple(dict_row[col] for col in group_by) if len(group_by) > 1 else dict_row[group_by[0]]
@@ -368,7 +379,7 @@ def group_list_by_columns(list_of_dicts: list, group_by: str, condense: bool = T
 #                     f" so we can add support to this case: insert error: {error}")
 # return duplicate_value, duplicate_column_name
 
-def generate_where_clause_for_ignore_duplicate(data_dict: dict, constraint_columns: list[list[str]]) -> (str, tuple):
+def generate_where_clause_for_ignore_duplicate(data_dict: dict, constraint_columns: list[list[str]]) -> Tuple[str, tuple]:
     where_clauses = []
     values = tuple()
 
@@ -397,16 +408,19 @@ def generate_where_clause_for_ignore_duplicate(data_dict: dict, constraint_colum
     return where, values
 
 
+# process field with dot in it
 def fix_select_clause_value(select_clause_value: str) -> str:
     """If some fields in select_clause_value have "." in them, we have to surround them with ``."""
     # Split the select clause value by commas to handle multiple fields
+    if select_clause_value is None:
+        return select_clause_value
     fields = select_clause_value.split(',')
 
     # Process each field
     fixed_fields = []
     for field in fields:
         field = field.strip()
-        if '.' in field:
+        if '.' in field and field[0] != '`' and field[-1] != '`':
             field = f'`{field}`'
         fixed_fields.append(field)
 
@@ -415,6 +429,25 @@ def fix_select_clause_value(select_clause_value: str) -> str:
     return fixed_select_clause_value
 
 
+def fix_where_clause_value(where_clause_value: str) -> str:
+    """If some fields in where_clause_value have "." in them, we have to surround them with ``."""
+    # Split the where clause value by AND or OR to handle multiple conditions
+    if where_clause_value is None:
+        return where_clause_value
+    conditions = where_clause_value.split()
+
+    # Process each condition
+    fixed_conditions = []
+    for condition in conditions:
+        condition_column = condition.split('=')[0].strip()  # Get the column name part of the condition
+        if '.' in condition_column and condition_column[0] != '`' and condition_column[-1] != '`':
+            condition = condition.replace(condition_column, f'`{condition_column}`')  # Surround the column name with `
+        fixed_conditions.append(condition)
+
+    # Join the fixed conditions back into a single string
+    fixed_where_clause_value = ' '.join(fixed_conditions)
+    return fixed_where_clause_value
+
 # TODO: run this with cache decorator if the above get_table_columns doesn't find anything.
 # def get_table_columns(schema_name: str = None, table_name: str = None) -> tuple:
 #     select_query = "SELECT column_name " \
@@ -431,4 +464,5 @@ def fix_select_clause_value(select_clause_value: str) -> str:
 #     self.logger.debug(object=locals())
 #     return columns_tuple
 
+
 module_wrapper(logger)
